
/////////////////////////////////////////////////////////////////
//
// File: bcd.c
//
// Author: J.F.Main
//
// Description:
//
//    Assembler code to convert an unsigned 32 bit long into a 10 digit ascii
//    (10 digits is the maximum for a 4 byte long variable).
//
//    It uses the minimum resources and occupys the least space (compared
//    to using built in routines that use 'long' multiply and divide.
//
//    This porvides a good example of integrating assembler into mikroC C code.
//
//    If you use the tivial solution in C by dividing by 10 and keeping
//    remainders then it uses the built in maths routines for multiplication
//    and division of long type variables which takes too much resource.
//
//    With built in routines approximately 400 bytes is used.
//    To get the smallest routine, machine code tricks have to be used
//    to convert the long to ascii.
//
//  Note:
//
//    The primary reason for using this is that if long maths are used
//     - divide or multiply then loads of code space is used up & it is
//    not possible to fit into 1k of memory (and do anything else useful).
//
// Version: 1.00 - Initial release.
//
// Copyright : Copyright  John Main 2006
//   http://www.best-microcontroller-projects.com
//   Free for non commercial use as long as this entire copyright notice
//   is included in source code and any other documentation.
//
/////////////////////////////////////////////////////////////////

/////////////////////////////////////////////////////////////////
//
// Notes on bank interaction
//
// If different banks are used then the machine code may not see
// variables.  This error can be caused by adding more code
// e.g. a debugging routine.
// When added if the 'total' code (variables) runs over more than
// on bank then PCLATH needs controlling (to access variables in the
// different banks).  Since the assembler has to be done manually
// it does not control PCLATH and so can not see variables and so fails.
// The solution is to fix assembler variables to a specific bank.
//
// Note symptom of error:
//   If tested alone, the code works perfectly using 496 bytes of memory
//
// If debugging code is added (uses up more memory) then the code
// consumes (in this case) 937 words of memory - the code stops.
//
// This could be seen in the debugger as bcd routine sometimes
// made a partial result or no result (depending changed C source i.e.
// while attempting to correct the problem by addding / removing code from main).
//
// Solution:
//   Reduce the code to the minimum that works.
//   Lock the variables used by the assembler routing
//      Look at the debugger variable locations and use 'absolute' directive.
//   NB also look at mikroC forum posts :
//      How can I access variables from within the ASM block?
//      (http://www.mikroe.com/forum/viewtopic.php?t=2042).
//      Problem integrating a C variable into ASM code
//      (http://www.mikroe.com/forum/viewtopic.php?t=2421&highlight=assembler)
//
/////////////////////////////////////////////////////////////////

// BCD digit store
static char DIG  absolute 0x30; // was array = faster if use assembler
static char DIG1 absolute 0x31; // was array = faster if use assembler
static char DIG2 absolute 0x32; // was array = faster if use assembler
static char DIG3 absolute 0x33; // was array = faster if use assembler
static char DIG4 absolute 0x34; // was array = faster if use assembler
static char DIG5 absolute 0x35; // was array = faster if use assembler
static char DIG6 absolute 0x36; // was array = faster if use assembler
static char DIG7 absolute 0x37; // was array = faster if use assembler
static char DIG8 absolute 0x38; // was array = faster if use assembler
static char DIG9 absolute 0x39; // was array = faster if use assembler

// Loop counters
static unsigned short BITCNT  absolute 0x3A; // bit counter.
static unsigned short DIGCNT  absolute 0x3B; // digit counter.

// Binary number store
static unsigned short BIN3    absolute 0x2C;
static unsigned short BIN2    absolute 0x2E;
static unsigned short BIN1    absolute 0x2D;
static unsigned short BIN0    absolute 0x2F;

//////////////////////////////////////////////////////////////////////
//
// Convert unsigned long 32bit (4byte) to unpacked BCD.
//
// Using serial hardware algorithm from application note xapp029.
//
// Identical to ulong2bcd except for BITCNT=16 changed from BITCNT=32
//
void ulong2bcd_p1(unsigned long num) {   // mod to 32 bits

unsigned short *p;
p = (unsigned short) &num;

// assignments made so assembler can see the C variables

bin0 = *(p++); bin1 = *(p++);
bin2 = *(p++); bin3 = *(p++);

dig  = 0; dig1 = 0; dig2 = 0;
dig3 = 0; dig4 = 0; dig5 = 0;
dig6 = 0; dig7 = 0; dig8 = 0; dig9 = 0;

DIGCNT = 0;   // keep these just before asm to intialise asm
BITCNT = 16;

// my asm
asm {

BITLOOP:
      movlw 10            // ; ten digits
      movwf bcd_DIGCNT
      movlw bcd_DIG       // ; location of output Digit
      movwf FSR           // ; set indirection reg to point at op digit.

      // Get the next bit
      rlf bcd_BIN0,F
      rlf bcd_BIN1,F
      rlf bcd_BIN2,F
      rlf bcd_BIN3,F      // ; MSBit shifted into the carry flag here

PROCESSDIGITSLOOP:
      rlf INDF,F        // Rotate carry into current digit
      movlw -10         // Get ready to compare to 10 decimal
      addwf INDF,W      // W = -10 + (dig[n] << 1)
      btfsc STATUS,C    // should be >=5 but rotated >=10  (c=1 if dig >=5).
      addlw 6           // greater than equal to 10 (5) so add 6 (3)
      addlw 10          // get rid of comparison test
      movwf INDF        // store so can test bit4.

      BCF STATUS,C
      btfsc INDF,4      // bit 4 is the overflow (effectively DC)
      BSF STATUS,C      // make C equal to this bit C=dig(bit4)

      andlw 0x0f        // keep upper nibble zero
      movwf INDF        // store the processed digit.

      INCF FSR,F        // next digit

      decfsz bcd_DIGCNT,F
      goto PROCESSDIGITSLOOP

      decfsz bcd_BITCNT,F
      goto BITLOOP

} // end asm

} // end long to bcd

//////////////////////////////////////////////////////////////////////
//
// ulong2bcd_p2
//
// Call this after calling ulong2bcd_p1 to complete the ulong to bcd
// conversion.
//
// Using these two routines lets you do work between bcd processing
// as the state of the bcd conversion is preserved after ulong2bcd_p1.
//
void ulong2bcd_process(void) {

BITCNT = 16;

   asm {
      call BITLOOP
   }
}

//////////////////////////////////////////////////////////////////////
//
// Convert unsigned long 32bit (4byte) to unpacked BCD.
//
void ulong2bcd(unsigned long num) {
unsigned short i;

   ulong2bcd_p1(num);

//   for (i=0;i<7;i++) {
      ulong2bcd_process(void);  // conversion 2nd part.
//   }
}

/////////////////////////////////////////////////////////////////
void packedBCDToStr(unsigned short packedBCD,unsigned short len, char *op) {
unsigned short *p=packedBCD, val, i;

   for (i=0;i<len;i++) {
      val  = (*p >> 4)  + '0';
      *(op++) = val;
      val  = (*p & 0x0f) + '0';
      *(op++) = val;
      p++;
   }
}

/////////////////////////////////////////////////////////////////
void unpackedBCDtoStr(unsigned short *bcd, unsigned short c, char *op) {
unsigned short i;

   for(i=0;i<c;i++) {
        *(op++) = (*(bcd++)) + '0';
   }
   *op=0; // terminate the string
}

/////////////////////////////////////////////////////////////////
// Reverse a string
// Argunents:
//    requires a null terminated string.
//
void strrev(char *s) {
unsigned short i,len;
char *e, *p, st;

   len = strlen(s);
   e   = s+len-1;

   for ( i=0; i<(len>>1); i++) { // >>1 = len/2
      st = *e; // temporary store.
      *e=*s;
      *s = st; // swapped.
      e--;
      s++;
   }
}

/////////////////////////////////////////////////////////////////
//
// mikroC long 4 bytes : 0 .. 4294967295
//
// This needs to take constant time - so does not break from loop
// always completes each digit.
//
void process_ulong2str(char *str) {

char *s=str;
unsigned short stop=0;

    // use s as mikroC does not initialise str
    unpackedBCDtoStr(&DIG, 10, str); // this null terminates.

    // Leading digit blanking (to space if value is ascii zero)
    strrev(str);
    s = str;
    while (*s) {
       if (*s!='0') { stop=1; }
       if (!stop) { // replace leading zeros with spaces.
          if (*s=='0') {*s=' '; }
       }
       s++;
    }
}

/////////////////////////////////////////////////////////////////
// mikroC long 4 bytes : 0 .. 4294967295
void ulong2str(long ip, char *str) {

char *s=str;

    ulong2bcd(ip);
    process_ulong2str(str);
}



